header {* Some addons to Isabelle/ZF *}

theory ZF_Addons
  imports Main_ZF
begin

text {* We need to add some additional lemmas to Isabelle/ZF before proceed. *}

(** Proof by Larry Paulson **)
lemma inj_inj_range: "f\<in>inj(A,B) \<Longrightarrow> f\<in>inj(A,range(f))"
  apply (unfold inj_def)
  apply (auto simp add: Pi_iff function_def)
done

(** Proof by Larry Paulson **)
lemma inj_bij_range: "f\<in>inj(A,B) \<Longrightarrow> f\<in>bij(A,range(f))"
  apply (auto simp add: bij_def inj_inj_range)
  apply (blast intro: inj_is_fun fun_is_surj elim:)
done

lemma range_subset: "f: A\<rightarrow>B \<Longrightarrow> range(f) \<subseteq> B"
proof -
  assume "f: A\<rightarrow>B" 
  hence "f\<in>Pow(A\<times>B)" by (simp add: Pi_def)
  hence "f \<subseteq> A\<times>B" by auto
  moreover
  hence "range(A\<times>B) \<subseteq> B" by auto
  ultimately show "range(f) \<subseteq> B" by auto
qed

lemma lam_is_fun: "f = (lam x:d. b(x)) \<Longrightarrow> f: d\<rightarrow>range(f)"
proof -
  assume eq: "f = (lam x:d. b(x))"
  have dom: "domain(f) = d" using domain_lam by (auto simp add: eq)
  with eq have "function(f)" using function_lam by auto
  moreover
  with eq have "relation(f)" using relation_lam by auto
  ultimately have "f: domain(f)\<rightarrow>range(f)" by (rule function_imp_Pi)
  with dom show ?thesis by auto
qed

lemma comp_fun2:
  "\<lbrakk> g: A\<rightarrow>B1; f: B0\<rightarrow>C; B1\<subseteq>B0 \<rbrakk> \<Longrightarrow> (f O g) : A\<rightarrow>C"
proof -
  assume "g: A\<rightarrow>B1" "f: B0\<rightarrow>C" "B1\<subseteq>B0"
  with `g: A\<rightarrow>B1` have "g: A\<rightarrow>B0" by auto (rule fun_weaken_type)
  from `g: A\<rightarrow>B0` `f: B0\<rightarrow>C` show "(f O g) : A\<rightarrow>C" using comp_fun by auto
qed

text {*
  This lemma superceedes the lemma @{term comp_eq_id_iff} in Isabelle/ZF:
*}
lemma comp_eq_id_iff1:
  "\<lbrakk> g: B\<rightarrow>A; f: A\<rightarrow>C \<rbrakk> \<Longrightarrow> (\<forall>y\<in>B. f`(g`y) = y) \<longleftrightarrow> f O g = id(B)"
proof -
  assume "g: B\<rightarrow>A" "f: A\<rightarrow>C"
  hence "f O g: B\<rightarrow>C" by (rule comp_fun)
  moreover
  have "id(B): B\<rightarrow>B" by (rule id_type)
  ultimately have m: "(\<forall>y\<in>B. (f O g)`y = id(B)`y) \<longleftrightarrow> f O g = id(B)" by (rule fun_extension_iff)
  from `g: B\<rightarrow>A` have c [simp]: "\<forall>y\<in>B. (f O g)`y = f`(g`y)" by auto
  have i [simp]: "\<forall>y\<in>B. id(B)`y = y" by auto
  from m show "(\<forall>y\<in>B. f`(g`y) = y) \<longleftrightarrow> f O g = id(B)" by simp
qed

lemma right_comp_id_any: "r O id(C) = restrict(r,C)"
unfolding restrict_def by auto

end